
#include "DiffusionMap.h"
#include <amlpp/amlpp.h>
#include <DataMat.h>
#include <omp.h>

using namespace amlpp;

/**
 * This method returns the computational equivalent of epsilon.
 * That is, it returns a number e such that the computer can represent
 * no number that is smaller than e
 */
double eps(double val)
{
	double inverse_radix = 0.5, machine_precision = val, temp ;
	do{
		machine_precision *= inverse_radix;
		temp = val + machine_precision ;
	}while (temp != val);
	return (machine_precision*2);
}

/**
 * This method calculates the square root of a matrix defined as
 * follows:
 * Given a matrix M, this method returns a matrix A such that
 * A*A=M
 * This method was implemented using the schur decomposition technique
 * in order to ensure that complex roots were also obtained
 */
void  sqrtm(const RealMatrix &A,ComplexMatrix &sqrtA)
{

	int n = A.rows();
	ComplexSchur SCHUR(A);
	ComplexMatrix Q  = SCHUR.U();
	ComplexMatrix T = SCHUR.T();
	sqrtA = zeros(n,n);
	for(int j = 1 ; j <= n; ++j)
	{
		sqrtA(j,j) = sqrt(T(j,j));
		#pragma omp for
		for(int i = j-1; 0<i; --i)
		{
			Complex s = 0;
			for(int k = i+1 ; k <= j-1; ++k)
			{
				s += sqrtA(i,k)*sqrtA(k,j);
			}
			sqrtA(i,j) = (T(i,j) - s)/(sqrtA(i,i) + sqrtA(j,j));
		}
	}
	sqrtA = Q*sqrtA*ctrans(Q);
}

/**
 * Receives a matrix A and returns its pseudo inverse in
 * the pinvA variable
 */
void pinv(const RealMatrix &A,RealMatrix &pinvA)
{

	int m = A.cols();
	int n = A.rows();
	if(n > m)
	{
		pinv(trans(A), pinvA);
		pinvA = trans(pinvA);
	}
	else
	{
		RealMatrix U, V, S;
// 		disp(A) ;
		RealSVD SVD(A);
		U = SVD.U();
		V = SVD.V();
		S = SVD.S();
		int rank = 1;
		double maxEig = std::max(n,m)*eps(S(1,1) );
		while ((rank < n) && ( maxEig < S(rank,rank))) ++rank;

		if (rank == 1) {
			pinvA  = zeros(n, m);
		}
		else
		{
			RealMatrix rankS = S(colon(1,rank), colon(1,rank));
			for (int i = 1; i <= rank; i++)
			{
				if(rankS(i, i)!= 0)
				{
					rankS(i, i) = 1.0/rankS(i,i);
				}
			}
			pinvA = V(colon(), colon(1,rank)) * rankS * trans(U(colon(), colon(1,rank)));
		}
	}
}

/**
 * A method that calculates Gaussian affinities between pixels and returns an affinity
 * matrix for an image using the following formula
 *
 * exp((-|xi-xj|^2)/2*(sigma_r^2))
 *
 * On all two pairs of pixels
 */
void euclidean_exp(const RealMatrix &X1, const RealMatrix &X2, RealMatrix &M, double sigma_r)
{
	int d = X1.rows();
	int n1 = X1.cols();
	int n2 = X2.cols();
	double temp, s, sigma = 1.0/sigma_r;
//	sigma = sigma*sigma;

	int j, i, k;
	double* M_ptr =  M.ptr();
	double* X1_ptr ;
	double* X2_ptr = X2.ptr();
	double* X1_ptr_temp ;
	double* X2_ptr_temp = X2.ptr();

	for(j = 1; j <= n2; ++j)
	{
		X1_ptr = X1.ptr();
		#pragma omp for
		for (i = 1; i <= n1; ++i)
		{
			s = 0;
			X1_ptr_temp = X1_ptr;
			X2_ptr_temp = X2_ptr;
			for (k =1; k <= d; ++k)
			{
				temp =  ((*(X1_ptr_temp))-(*(X2_ptr_temp)));
				s+=temp*temp;
				++X1_ptr_temp;
				++X2_ptr_temp;
			}
			X1_ptr+=d;
			*M_ptr++ =  exp(-sigma*s);
		}
		X2_ptr+=d;
	}
}

/**
 * Assist method that receives an image of 'height' by 'width' pixels
 * and returns a matrix of ('height')x('width')x(number of channels) in which
 * the original pixels are laid out in order and divided into the images
 * channels
 */ 
void patchToPoints(std::vector<float *>  &img,int height, int width ,RealMatrix & mat)
{
	int ind1, ind2 = 0;
	for(int y = 0 ; y < width ;++y)
	{
	    #pragma omp for
		for(int x = 0 ; x < height ;++x)
		{
			ind1 = y*height+x+1;
			ind2 = x*width+y;
			mat(ind1,1)= img[0][ind2];
			mat(ind1,2)= img[1][ind2];
			mat(ind1,3)= img[2][ind2];
		}
	}
}

/**
 * ibid
 */
void patchToPoints_full(std::vector<float *>  &img, int height, int width ,RealMatrix & mat, double sc)
{
	int ind1, ind2= 0;
	for(int y = 0 ; y < width ;++y)
	{
	    #pragma omp for
		for(int x = 0 ; x < height ;++x)
		{
			ind1 = y*height+x+1;
			ind2 = x*width+y;
			mat(ind1,1)= img[0][ind2];
			mat(ind1,2)= img[1][ind2];
			mat(ind1,3)= img[2][ind2];
			mat(ind1,4)= sc*((double)y/(double)width);
			mat(ind1,5)= sc*((double)x/(double)height);
		}
	}
}


/**
 * This method returns a set of eigenvalues and eigenvectors from an image's
 * affinity matrix (that together make up the image's diffusion maps) given
 * the image (with its dimensions) an affinity parameter,
 * a number of pixels to sample for an affinity matrix and an expected number
 * of eigenvalues/vectors
 * Input:
 * 	parameter: img - A vector that represents an image by channels
 * 	parameter: height - The height of the image
 * 	parameter: width - the width of the image
 * 	parameter: sigma_r - the sigma parameter for the Gaussian affinity equation
 * 	parameter: nvec - The number of eigen values to be calculated
 * 	parameter: nsamples - The number of vectors in the affinity matrix to be sampled for estimating the diffusion
 * 						maps (number of vectors in affinity matrix is number of pixels in original image)
 * 	parameter: sc A component that allows us to relate to the distance between pixels in addition to the
 *            'color-distance'. Must be between [0,1] (error checking left up to user)
 *
 * Output:
 * return value: eigenvalues The eigenvalues of the images affinity matrix
 * return value: diffusion_map The diffusion map estimate
 */
void decompose( std::vector<float *> &img,int height, int width, double sigma_r,  int nvec, int nsamples ,double sc, std::vector<float> &eigenvalues , std::vector<DataMat> &diffusion_map)
{
/* 
	diffusion_map.resize(nvec);*/
	std::cout << "Using " << nvec << " dimensions, " << nsamples << " samples, " << sigma_r << " as sig_r" << endl ;
	RealMatrix mat( height*width, img.size()+((sc==0)?0:2));
	if(sc == 0)
	{
		patchToPoints(img,height, width, mat);
	}
	else
	{
		patchToPoints_full(img,height, width, mat,sc);
	}
	int size = mat.rows();
	
	//  here we create two sets of indices to cover all of the matrix

	IntegerRowVector ind = randperm(size);
	RealMatrix Ximp( mat.cols(), nsamples);
	RealMatrix Xoth( mat.cols(), size - nsamples);

	// Here the indices are mixed and matched for the sparse A matrix and  B matrix
	for(int j = 1 ; j <= Ximp.cols() ; ++j)
	{
		for(int z= 1 ; z <= Ximp.rows(); ++z)
		{
			(Ximp)(z,j)=  mat(ind(j),z);
		}
	}
	for(int j = 1 , i = Ximp.cols()+1; j <= Xoth.cols() ; ++j,++i)
	{
		for(int z= 1 ; z <= Xoth.rows(); ++z)
		{
			(Xoth)(z,j)=  mat(ind(i),z);
		}
	}
	mat.~RealMatrix();

	RealMatrix A(Ximp.cols(), Ximp.cols());
	RealMatrix B(Ximp.cols(), Xoth.cols());

	// And finally, we calculate the distances between points on the A matrix
	// and between points on the A matrix and points on the B matrix
	// Recal that eventually we want to build:
	//
	//                               A | B'
	//                               -----
	//                               B | C

	euclidean_exp(Ximp,Ximp,A, sigma_r);
	euclidean_exp(Ximp,Xoth,B, sigma_r);
	Ximp.~RealMatrix();
	Xoth.~RealMatrix();
	// Nystrom
	// Sets the upper part of the Nystrom matrix
	// As advertised
	//                   A | B'
	RealMatrix Up = vertcat(A, trans(B));

	// d1 shall be the a vector of the sum of all the columns of the
	// Up matrix

	RealMatrix d1 = sum(Up,1);

	RealMatrix pinvA;	
	pinv(A, pinvA);
	RealMatrix d2 = sum(B,1)+trans(sum(B,2))*pinvA*(B);
	RealMatrix dhat(d2.cols()+d1.cols(), 1);
	dhat= trans(sqrt(amlpp::div(1,horzcat(d1, d2))));

	d1.~RealMatrix();
	d2.~RealMatrix();

	int Arows = A.rows();
	int Acols = A.cols();
	double* A_ptr =  A.ptr();
	double* dhat_ptr_i =  dhat.ptr();
	for(int i = 1 ; i <= Acols; ++i)
	{
		double* dhat_ptr_j =  dhat.ptr();
		#pragma omp for
		for(int j = 1 ; j <= Arows; ++j)
		{
			(*A_ptr) *= (*dhat_ptr_j)*(*dhat_ptr_i);
			++A_ptr;
			++dhat_ptr_j;
		}
		++dhat_ptr_i;
	}
	int Brows = B.rows();
	int Bcols = B.cols();
	double* B_ptr =  B.ptr();
	dhat_ptr_i =  dhat.ptr()+Brows;
	for(int i = 1; i <= Bcols; ++i )
	{
		double* dhat_ptr_j =  dhat.ptr();
		#pragma omp for
		for(int j = 1 ; j <= Brows; ++j)
		{			
			(*B_ptr) *= (*dhat_ptr_j)*(*dhat_ptr_i);
			++B_ptr;
			++dhat_ptr_j;
		}
		++dhat_ptr_i;
	}
	ComplexMatrix Asi ;
	pinv(A, pinvA);
	sqrtm(pinvA,Asi);
	pinvA.~RealMatrix();
	ComplexMatrix Q = A+Asi*(ComplexMatrix((B)*trans(B))*Asi);
	A.~RealMatrix();
	B.~RealMatrix();
	ComplexMatrix L, U, T;	
	ComplexSVD SVD(Q);
	U = SVD.U();
	T = SVD.V();
	L = SVD.S();
	Q.~ComplexMatrix();
	SVD.~ComplexSVD();
	ComplexMatrix D = L.diag();
	eigenvalues = std::vector<float>(nvec);
	for(int i = 1 ; i < nvec+1 ; ++i )
	{
		eigenvalues[i-1] = D(i+1,1).real();
	}
	ComplexMatrix V = ComplexMatrix(Up)*(Asi*(U*ComplexMatrix(bdiag(div(ones(D.rows(),D.cols()),sqrt(abs(D)))))));
	Up.~RealMatrix();
	RealMatrix E (size, nvec);
	disp(D(colon(1,nvec+1))) ;

	ComplexMatrix res;
	for (int i=2; i <= nvec+1; ++i){

		res = div(V(colon(), colon(i,i)),V(colon(), colon(1,1)));
		#pragma omp for
		for(int j = 1 ; j <= size ; ++j)
		{
			E(ind(j),i-1) = res(j,1).real();
		}
//		diffusion_map[i-2] = new float[ width * height ];
		for(int h = 0, Eindex = 1; h <  width; ++h )
		{
			for(int w = 0 ; w <  height; ++w ,++Eindex)
			{

// 			    diffusion_map[i-2][w*width+h] = E(Eindex,i-1);
			    diffusion_map[i-2] (w,h) = E(Eindex,i-1);
			}
		}
	}
}
