/*
 * Accurate binary image selection from inaccurate user input
 *
 * Authors: Kartic Subr, Sylvain Paris, Cyril Soler, Jan Kautz
 *
* Copyright (2013):Kartic Subr. 
 * All rights reserved. 
 *
 * For details, see the paper:
 * Accurate binary image selection from inaccurate user input
 * Computer Graphics Forum 32, 2 (2013)
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for educational, research, and non-commercial
 * purposes, without fee and without a signed licensing agreement, is
 * hereby granted, provided that the above copyright notice and this
 * paragraph appear in all copies modifications, and distributions.
 *
 * Any commercial use or any redistribution of this software
 * requires a license.
 *
 * For further details, contact Kartic Subr (kartic@gmail.com)
 */

#include <ostream>
#include <fstream> 
#include <cmath> 
 
#include <PFMImage.h>
#include <DataMat.h>
#include <FactoredImage.h>

#include <Eigen/Eigenvalues> 
#include <Eigen/Dense>

#include <map>
#include <omp.h>

#include <DiffusionMap.h>

using namespace std;
using namespace Magick; 
// using namespace boost;
using namespace Eigen ;
void knn(float* ref_host, int ref_width, float* query_host, int query_width, int height, int k, float* dist_host);
void knn(float* ref_host, int ref_width, float* query_host, int query_width, int height, int k, float* dist_host, int* ind_host);
  
namespace
{
    inline MatrixXd Tau(const MatrixXd& M)
    {
	MatrixXd H ;
	H.setIdentity(M.rows(), M.cols()) ;
	const double d = 1.0/M.rows() ;
	for(int i(0); i<M.rows(); i++)
	    for(int j(0); j<M.cols(); j++)
		H(i,j) = H(i,j) - d ;
	return -.5*H*M*H.transpose();
// 	return M;
    }

    inline void IDToRC(int id, int w, int h, int& r, int& c)
    {
	    r = id%h ;
	    c = id/h ;
    }

    inline int ToID(int i, int j, int w, int h)
    {
    //     return i*w + j ;
	return j*h + i ;
    }
}

ostream& operator<<(ostream& os, const Params& p) 
{
    os << "Downsampling= " << p.ds << endl
	    << "NystromFrac= " << p.NystromFrac << endl 
	    << "SigR= " << p.sigr << endl 
	    << "SigS= " << p.sigs << endl 
	    << "nDim= " << p.nDim << endl 
	    << "UseScribbles= " << p.UseScribbles << endl 
	    << "RunShortestPaths= " << p.UseShortestPaths << endl 
	    << "distType= " << p.distTYPE << endl 
	    << "UseLUV= " << p.LUVSpace << endl
	    << "MaxNN= " << p.kNN << endl
	    << "PrecomputeNN= " << p.PrecomputeNN << endl
	    << "NN_MaxRad= " << p.knnRad << endl 
	    << "TileXNN= " << p.nTileX << endl
	    << "UseXYforNN= " << p.UseXYDims << endl 
	    <<  "FilterRadius= " << p.FilterRadius << endl 
	    <<  "PatchMetricWSize= " << p.PatchMetricWSize << endl ;
    return os; 
}


istream& operator>>(istream& is, Params& p) 
{
    string dummy;
    int tmp ;
        is >> dummy >> p.ds 
	    >> dummy >> p.NystromFrac  
	    >> dummy >> p.sigr  
	    >> dummy >> p.sigs  
	    >> dummy >> p.nDim  
	    >> dummy >> p.UseScribbles  
	    >> dummy >> p.UseShortestPaths  
	    >> dummy >> tmp
	    >> dummy >> p.LUVSpace 
	    >> dummy >> p.kNN 
	    >> dummy >> p.PrecomputeNN 
	    >> dummy >> p.knnRad  
	    >> dummy >> p.nTileX 
	    >> dummy >> p.UseXYDims  
	    >> dummy >> p.FilterRadius 
	    >> dummy >> p.PatchMetricWSize ;
	    
	p.distTYPE = int2DistMetric[tmp] ;   
	
    return is ;
} 
   
FactoredImage::~FactoredImage() 
	{ 
	    if (_AllocatedPI) 
		delete _pI ; 
// // 	    if (_params.Verbosity>1) cout << " Deleting NN object" << endl ;
// 	    if (_params.PrecomputeNN && _pNN) 
// 		delete _pNN; 
// // 	if (_params.Verbosity>1) cout << "Deleting PC object" << _pPC << endl ;if (_pPC!=NULL) delete _pPC;  if (_params.Verbosity>1) cout << endl; 	    
	}

void FactoredImage::ComputeMinMaxDims()  
{
        const int p(_O.size()) ;

     vector<double> mins(p, 99999), maxs(p,-99999) ;
    _minDims.resize(p) ;    _maxDims.resize(p) ;
    for (int j(0); j<p; j++) {	_minDims[j] = 99999 ; _maxDims[j] = -99999 ;    }
    
    double minel, maxel ;
    for (int row(0); row<_h; row++)
    {
	for (int col(0); col<_w; col++)
	{
	    for(int j(0); j<p; j++) 
	    {
		const double rhs(_O[j](row, col));	    
		minel = min(rhs, minel) ;
		maxel = max(rhs, maxel) ;
		mins[j] = min(mins[j], rhs) ;
		maxs[j] = max(maxs[j], rhs) ;		
		_minDims[j] = min(_minDims[j], rhs) ;
		_maxDims[j] = max(_maxDims[j], rhs) ;
	    }
	}
    }
    _minCoord = minel ;
    _maxCoord = maxel ;
    double maxDiam (0) ;
    for (int j(0); j<p; j++) 
    {
	maxDiam = max(maxDiam, maxs[j]-mins[j]) ;
    }
    _maxDiam = maxDiam ;
    if (_params.Verbosity>1) cout << "\t\t Min el = " << minel <<  " Max el = " << maxel << endl ;
}

FactoredImage::FactoredImage(const Params& param, const PFMImage& I) : _pI(&I),  _w(I.width), _h(I.height), _params(param), 
// 													_pNN(NULL),
											      _DH(I, param.distTYPE, param.sigr, param.sigs, param.LUVSpace, param.PatchMetricWSize), _AllocatedPI(false)
{
	_pI = new PFMImage(I) ;
	_AllocatedPI = true; 
	vector<int> RowIndices, ColIndices ;
	vector<double> Vals ;
	int sz(_w*_h); 
	int c = param.NystromFrac * param.NystromFrac * sz ;
	c = max(param.nDim+1, c) ;
	if (_params.Verbosity>1) cout << "Size of dist matrix: " << sz << "x" << sz << endl ;
	DataMat C, A ;
	vector<int>  P ;

	if (param.distTYPE==DMAP || param.distTYPE==BIH)
	{
	    ComputeDiffusionMap(param) ;
	}
	else
	{
	    if (param.UseShortestPaths)
	    {
		// For IsoMap
		// 		ComputeSparseDistanceMatrix(RowIndices, ColIndices, Vals, sz) ;
		// 		NystromShortestPaths(Vals, RowIndices, ColIndices, C, A, P, c, sz);
	    }
	    else
	    {
		// For LMDS
		NystromAllPairs(C, A, P, c, sz);
	    }
		
	    Factorize(C,A,P, sz, c) ;
	}
	
	ForceRGBWithoutEmbedding() ;
}


void FactoredImage::ComputeDiffusionMap(const Params& param)
{
    int p = _params.nDim ;
    _O.clear() ;
    DataMat Dummy(_h,_w) ;    
    for (int i=0; i<p; i++) _O.push_back(Dummy) ;
    vector<float*> img(3) ;
    for (int i(0); i<3; i++) img[i]=new float [_w*_h] ;
    
    const PFMImage& I = *_pI ;
    double MaxL (0) ;
    double MinL (0) ;

    // Warning! Bad variable nomenclature  for x and y (from the original diffusion maps implementation)    
	for(int y = 0 ; y < _w;++y)
	{
		for(int x = 0 ; x < _h ;++x)
		{			
			int ind2 = x*_w+y;
			PFMPixel p = I(x,y) ;
			if (param.LUVSpace)
			{
			  p = RGB2LUV(p) ;
			  MaxL = max(MaxL, (double)p.r) ;
			  MinL = min(MinL, (double)p.r) ;

			}
			
			img[0][ind2] = p.r ;
			img[1][ind2] = p.g ;
			img[2][ind2] = p.b ;
// 			cout << " " << I(x,y).r  << flush ;
		}
	}
	
	if (param.LUVSpace)
	{
	    cout << " Adjusting DR of L " << endl ;
	  	 double DivLdr(1.0/(MaxL-MinL)) ;
	 if (fabs(MaxL-MinL) <1e-8) DivLdr = 1.0 ;
	 
	 for(int y = 0 ; y < _w;++y)	
		for(int x = 0 ; x < _h ;++x)
		{
		  	int ind2 = x*_w+y;
			PFMPixel p = RGB2LUV(I(x,y)) ;			
			p = DivLdr*(p-PFMPixel(MinL,MinL,MinL)) ;
			img[0][ind2] = p.r ;
			img[1][ind2] = p.g ;
			img[2][ind2] = p.b ;
		}		    
	}
	
    double sigma_r = param.sigr ;
    int c = max(p+1.0, param.NystromFrac * param.NystromFrac * _w*_h) ;
    vector<float> eigenvalues ;
    cout << " Creating DiffusionMap using " << c << " samples " << endl ;
    decompose( img,_h, _w, sigma_r,  p, c ,0, eigenvalues , _O) ;
    p = eigenvalues.size() ;    
    const double eps2(1e-9) ;
    double eps1(.01) ;
    for(int i(0); i<p; i++)
    {
	if (param.distTYPE==DMAP)
	    _O[i] *= pow(eigenvalues[i], param.sigs) ;
	else if (param.distTYPE==BIH)
	{
	    if (eigenvalues[i]<1e-5) 
		eps1=.01; 
	    else 
		eps1=0.0;
	    _O[i] *= -1/(log(eigenvalues[i] + eps1)+eps2) ;    
	}
    }
    for (int i(0); i<3; i++) delete [] img[i];
    
    ComputeMinMaxDims()  ;
}

vector<double> FactoredImage::operator()(int i, int j) const 
{
    vector<double> v(_O.size(),0);
    for (int k(0); k<_O.size(); k++) v[k] = _O[k](i,j) ;
    return v; 
}
 
  

//  Rather than use shortest paths, directly compute distances for Nystrom approximation
void FactoredImage::NystromAllPairs( DataMat& C, DataMat& A, vector<int>& P, const int c, const int n) 
{
     
    // Pad C and A matrices to be cxc and cxn
    C.resize(c, c) ;
    A.resize(c,n-c) ;
   
    // Sample c out of n
    P.resize(n) ;

    if (_params.Verbosity>1) cout << " ... Sampling image" << endl ;
    NystromRandomSample(P,c,n) ;

    if (_params.Verbosity>1) cout << " ... Filling C and A matrices " << c << " " << n << endl ;
    #pragma omp parallel for
    for (int i=0; i<c; i++)
    {
	if (!(i%100)) if (_params.Verbosity>1) cout << i << " of " << c << endl ; 
	for (int j(0); j<n; j++) 
	{
	    if(i==j) continue ;
	    int ix = P[i] ;
	    int jx = P[j] ;
	    int ri, ci, rj, cj ;
	    IDToRC(ix, _w, _h, ri, ci);
	    IDToRC(jx, _w, _h, rj, cj);

	    double dist = _DH.getDistance(ri, ci, rj, cj) ; 
	    
	    if(j<c)
	    {
		C(i,j)= dist ;
	    }
	    else
	    {
		A(i,j-c) = dist ;
	    }
	}
    }       
    if (_params.Verbosity>1) cout << "             Done             " << endl ;
}

void FactoredImage::NystromRandomSample(vector<int>& P, int c, int n)
{
    //     Simple random sampling
        for(int i(0); i<n; i++) P[i] = i ;
        random_shuffle(P.begin(), P.end());
}


void FactoredImage::Factorize(const DataMat& C, const DataMat& A, const vector<int>& P, const int n, const int c)
{
    const int p = _params.nDim ;
    
//     if (_params.Verbosity>1) cout << "\t Constructing C, A matrices" << endl ;
    MatrixXd Cm (c,c);
    MatrixXd Bm (c,n);
    #pragma omp parallel for
    for(int i=0; i<c; i++)
	for(int j(0); j<c; j++)
	    Cm(i,j)  = C(i,j) * C(i,j) ;
    
    VectorXd Davg (c) ;
    #pragma omp parallel for
    for(int i=0; i<c; i++)
    {
	Davg[i] = 0.0 ;	
	for(int j(0); j<n; j++)
	{
	    if(j<c)
		Bm(i,j) = C(i,j) * C(i,j);
	    else
		Bm(i,j) = A(i,j-c) * A(i,j-c);

// 	    Davg[i] += Bm(i,j) *Bm(i,j);
	    Davg[i] += Bm(i,j);
	}
    }
    Davg = 1.0/n * Davg ;
    if (_params.Verbosity>1) cout << ".. Eigendecomposition in progress  " << endl ;
    Cm = (Cm + Cm.transpose()) /2 ;
    SelfAdjointEigenSolver<MatrixXd> es(Tau(Cm));
    VectorXd  lambda = es.eigenvalues();
    if (_params.Verbosity>1) cout << "'\t\t" << Transpose<VectorXd>(lambda) << endl ;
    const int nEvals(es.eigenvalues().size()) ;
    _Eigvals.resize(nEvals); 
    for(int i(0); i<nEvals;i++)
    {
	_Eigvals[i] = (es.eigenvalues()(i)) ;
    }
    
    
    MatrixXd L(p, c) ;
    #pragma omp parallel for
    for(int i=0; i<p; i++)
	for(int j(0); j<c; j++) 
	{
	    if(lambda[c-p+i] < 1e-3)
		L(i,j)=0 ;
	    else
		L(i,j) = es.eigenvectors().col(c-p+i)[j] / sqrt(lambda[c-p+i]) ;
	    
	}
	
    _O.clear() ;
    DataMat Dummy(_h,_w) ;
    
    for (int i=0; i<p; i++) _O.push_back(Dummy) ;
    
    if (_params.Verbosity>2) cout << "\t Computing coordinates ..." << endl ;
    double minel(9999), maxel(-9999) ;

    vector<double> mins(p, 99999), maxs(p,-99999) ;
    _minDims.resize(p) ;
    _maxDims.resize(p) ;
    for (int j(0); j<p; j++)
    {
	_minDims[j] = 99999 ;
	_maxDims[j] = -99999 ;
    }
    
    for (int i(0); i<n; i++)
    {
	VectorXd Di = Bm.col(i) ;
	VectorXd Xi = -.5 *L*(Di-Davg) ;
	
	const int row = P[i]%_h ;
	const int col = floor(P[i]/_h) ;
	for (int j(0); j<p; j++)
	{
	    const double rhs(Xi[j]) ;
	    _O[j](row, col) =  rhs;
	    
	    minel = min(rhs, minel) ;
	    maxel = max(rhs, maxel) ;
	    mins[j] = min(mins[j], rhs) ;
	    maxs[j] = max(maxs[j], rhs) ;
	    
	    _minDims[j] = min(_minDims[j], rhs) ;
	    _maxDims[j] = max(_maxDims[j], rhs) ;
	}
    }
    _minCoord = minel ;
    _maxCoord = maxel ;
    double maxDiam (0) ;
    for (int j(0); j<p; j++) 
    {
	maxDiam = max(maxDiam, maxs[j]-mins[j]) ;
    }
    _maxDiam = maxDiam ;
    if (_params.Verbosity>2) cout << "\t\t Min el = " << minel <<  " Max el = " << maxel << endl ;
    
}

void FactoredImage::Normalize()
{
    if (_params.Verbosity>1) cout << " Normalizing factored image ... " << endl ;
    for (int i(0); i<_O.size(); i++)
    {
	_O[i] -= _minCoord;
	_O[i] *= 1.0/(_maxCoord-_minCoord) ;
    }
}

double FactoredImage::SqDistBetweenPixels(int r1, int c1, int r2, int c2) const 
{
    double d(0) ;
    for (int i(0); i<_O.size(); i++)
    {
	const double d1 = _O[i](r1,c1) ;
	const double d2 = _O[i](r2,c2) ;
	d += (d2-d1) * (d2-d1) ;
    }   
    return d ;
}


void FactoredImage::ForceRGBWithoutEmbedding() 
{
    const int nd(_O.size()) ;
    if (_params.distTYPE!=EUC) return ;
    if (nd<3) return; 
    
    const int h(_O[0].NRows()), w(_O[1].NCols()) ;
    cout << "====================== Forcing EUC " << nd << " "<< w << " " << h << endl ;
    for(int i(0); i<h; i++)
	for(int j(0); j<w; j++)
	{
// 	    cout <<  " " << i <<  " " << j << " --- " << h << " " << w  << flush << _O[0](i,j) << endl ;;
	    PFMPixel p =(*_pI)(i,j) ;
	    _O[0](i,j) = p.r ;
	    _O[1](i,j) = p.g ;
	    _O[2](i,j) = p.b ;
	}
}
