/*
 * Accurate binary image selection from inaccurate user input
 *
 * Authors: Kartic Subr, Sylvain Paris, Cyril Soler, Jan Kautz
 *
 * Copyright (2013):Kartic Subr. 
 * All rights reserved. 
 *
 * For details, see the paper:
 * Accurate binary image selection from inaccurate user input
 * Computer Graphics Forum 32, 2 (2013)
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for educational, research, and non-commercial
 * purposes, without fee and without a signed licensing agreement, is
 * hereby granted, provided that the above copyright notice and this
 * paragraph appear in all copies modifications, and distributions.
 *
 * Any commercial use or any redistribution of this software
 * requires a license.
 *
 * For further details, contact Kartic Subr (kartic@gmail.com)
 */

#include <Hist.h>
#include <iostream>
#include <cassert>
#include <iomanip>
#include <opencv2/imgproc/imgproc.hpp>

using namespace cv;
using namespace std;

Histogram::Histogram(int nBins, double a, double b) : _a(a), _b(b), _n(nBins), _initd(false) 
{     Init(_n, _a, _b) ; }

Histogram::Histogram() : _initd(false), _h(0)
{}

void Histogram::Init(int nBins, double a, double b)
{    
    _n=nBins ; 
    _a = a;
    _b = b; 
    _h = gsl_histogram_alloc (nBins); 
    _initd=true; 
    gsl_histogram_set_ranges_uniform (_h, a, b);
}

Histogram::~Histogram() 
{    gsl_histogram_free (_h); }

void Histogram::Add(double elem) 
{    gsl_histogram_increment (_h, elem); }


double Histogram::GetFreq(int i) const 
{
    return gsl_histogram_get(_h, i) ;
//     if (i>_n) return -1 ;
//     return _h->bin[i] ;
}

int Histogram::NBins() const 
{    return _n ;}


MultivalHist::MultivalHist(int nBins, double min, double max, int nVals):_n(nBins), _a(min), _b(max), _nV(nVals), _inited(false)
{
	Init(nBins, min, max, _nV) ;
}

MultivalHist::MultivalHist(): _inited(false)
{
}

MultivalHist::~MultivalHist() 
{
//     cout << " Destroying MultivalHist " << _inited << endl ;
    if (_inited)
	for(int i(0); i<_h.size(); i++) gsl_histogram_free (_h[i]);
}
	
void MultivalHist::Init(int nBins, double a, double b, int nVals) 
{
    _a=a; 
    _b=b; 
    _n = nBins; 
    _nV = nVals;     
    for(int i(0); i<_nV; i++)
    {
// 	cout << i << endl ;
	_h.push_back(gsl_histogram_alloc (nBins));
// 	cout << _h[i] << endl ;
	gsl_histogram_set_ranges_uniform (_h[i], a, b);
	
    }
    _inited = true ;
}
	

void MultivalHist::Add(const vector<double>& multielems) 
{
    assert(multielems.size()==_nV) ;
    assert(_h.size()==_nV) ;
    if (_h.size()!=_nV) cout << " Warning !!  Size mismatch " << endl ;
    for (int i(0); i<_nV; i++)
    {
// 	cout << " " << multielems[i] << " === " << flush ;
	const double& x(min(1-1e-8, multielems[i]));
		
	size_t bb;
// 	cout << x << " " << flush ;
	int b = gsl_histogram_find (_h[i], x, &bb) ;
	
	
	double r(_h[i]->range[bb]) ;
	double dr(_h[i]->range[1]-_h[i]->range[0]);
	
// 	gsl_histogram_accumulate (_h[i], x, 1);
	
// 	cout << r << " " << dr << " " << x << " " << bb << endl ;
	if (bb==_n-1 || (bb==0 && x<dr/2.0) ) 
	{	    
	    gsl_histogram_accumulate (_h[i], x, 0.5);
	}	    
	else
	{
	    size_t b1(bb), b2(bb+1) ;
	    double w1(r-x+1.5*dr), w2(1-w1) ;
	    
	    if(x<r+dr/2)
	    {
		b1 = bb-1 ;
		b2 = bb ;
		swap(w1, w2) ; 
	    }
	    _h[i]->bin[b1]+=w1; 
	    _h[i]->bin[b2]+=w2; 
	}
    }
}

vector<double> MultivalHist::GetFreq(int bbin) const 
{
//     cout << bin << endl ;
    vector<double> ret(_nV,0) ;
    for (int i(0); i<_nV; i++) 
    {	
	ret[i] =  gsl_histogram_get(_h[i], bbin) ; 
    }

    return ret ;
}

vector<double> MultivalHist::GetvalHist(int channel) const 
{
    vector<double> ret(_n,0) ;
    for (int i(0); i<_n; i++) 
    {	
	ret[i] =  gsl_histogram_get(_h[channel], i) ; 
    }

    return ret ;
}

int MultivalHist::NBins() const 
{return _n;}

double Cosine(const MultivalHist& h1, const MultivalHist& h2) 
{
    const int n(h1.NBins()) ;
    const int nV(h1._nV) ;
    double dist(0) ; 
    
    vector<double> m1(3,0), n1(3,0), d2(3,0) ; 
    for (int i(0); i<n; i++)
    {
	const vector<double>& p = h1.GetFreq(i) ;
	const vector<double>& q = h2.GetFreq(i) ;
	
	for (int j(0); j<nV; j++)
	{
	    m1[j] += p[j]*p[j]; 
	    n1[j] += q[j]*q[j] ;
	    d2[j] += p[j]*q[j] ;
	}		
    }    
    
    	for (int j(0); j<nV; j++)
	{
	    double den(d2[j]);
	    if( m1[j]>1e-5 && n1[j]>1e-5) den = 1.0 / (sqrt(m1[j]*n1[j])) ;
	    dist += acos(d2[j]*den) ;
// 	    if (isnan(dist)) 
// 	    {
// 		for (int i(0); i<n; i++)
// 		{
// 		    const vector<double>& p = h1.GetFreq(i) ;
// 		    const vector<double>& q = h2.GetFreq(i) ;
// 		    cout << " ---> "  << p[0] << " " << q[0] << " " << flush ;
// 		}
// 		
// 		cout << endl << d2[j] << " " << den  << endl ;
// 		cin.ignore(); 
// 	    }
	}
    return dist ;
}


double ChiSqOld(const MultivalHist& h1, const MultivalHist& h2) 
{
    const int n(h1.NBins()) ;
    const int nV(h1._nV) ;
    double dist(0) ; 
    
    for (int i(0); i<n; i++)
    {
        const vector<double>& p = h1.GetFreq(i) ;
        const vector<double>& q = h2.GetFreq(i) ;

        for (int j(0); j<nV; j++)
        {
            double den(1.0) ;
            if (fabs(p[j]+q[j]) > 1e-4)  den = 1.0/(p[j]+q[j]) ;
            dist += pow(p[j] - q[j], 2.0)*den   ;
        }
    }    
    return dist ;
}

double EMDdistOld(const MultivalHist& h1, const MultivalHist& h2) 
{
    const int n(h1.NBins()) ;
    const int nV(h1._nV) ;
    double dist(0) ; 
    vector<float> dummy(n+1,1) ;
    vector<vector<float> > v1(nV, dummy) , v2(nV, dummy) ;
    for (int i(0); i<n; i++)
    {
        const vector<double>& p = h1.GetFreq(i) ;
        const vector<double>& q = h2.GetFreq(i) ;

        for (int j(0); j<nV; j++)
        {
            v1[j][i+1] = p[i] ;
            v2[j][i+1] = q[i] ;
        }
    }        
    for (int j(0); j<nV; j++)
    {
        dist += EMD(v1[j], v2[j], CV_DIST_L2) ;
    }
    return dist ;
}
