/*
 * Accurate binary image selection from inaccurate user input
 *
 * Authors: Kartic Subr, Sylvain Paris, Cyril Soler, Jan Kautz
 *
 * Copyright (2013):Kartic Subr. 
 * All rights reserved. 
 *
 * For details, see the paper:
 * Accurate binary image selection from inaccurate user input
 * Computer Graphics Forum 32, 2 (2013)
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for educational, research, and non-commercial
 * purposes, without fee and without a signed licensing agreement, is
 * hereby granted, provided that the above copyright notice and this
 * paragraph appear in all copies modifications, and distributions.
 *
 * Any commercial use or any redistribution of this software
 * requires a license.
 *
 * For further details, contact Kartic Subr (kartic@gmail.com)
 */

#include <tclap/CmdLine.h>
#include<iostream>
#include<FactoredImage.h>
#include<PFMImage.h>
#include <Magick++.h> 
#include <string>
#include <algorithm>

#include <iostream>
using namespace std;
using namespace TCLAP ;
using namespace Magick; 

void CallCRFVersion(int argc, char* argv[], const PFMImage& I, FactoredImage* pFI, const string& fg, const string& bg, const string& o, int ds, double sigma, double wt, bool distComp)  ;
string IMPATH  ;


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//     Convert to internal image datastructure
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Magick2PFM( Image& I, PFMImage& J) 
{
     // Allocate pixel view 
    Pixels view(I);
    
     size_t columns = I.columns() ;
     size_t rows = I.rows() ; 
     J.Resize(columns, rows) ;
     const PixelPacket *pixels = view.getConst(0,0,columns,rows);
     for ( ssize_t row = 0; row < rows ; ++row )
         for ( ssize_t column = 0; column < columns ; ++column )
	 {
	     Color c2 = *pixels++ ; 
	     ColorRGB c = c2;
	     PFMPixel p(c.red(), c.green(), c.blue()) ;
	     J(row, column) = p;
	 }
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//     Read input image
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void ReadImage(const string& ImgIn, PFMImage& I, int ds)
{
   
    Image Unmarked;
    string unmkfile = IMPATH + ImgIn ;
      try 
    { 
	Unmarked.read( unmkfile );
    }
    catch( Exception &error_ ) 
    {
	throw error_ ;
    }
    
//     Downsample image
    int x = Unmarked.columns() ;
    int y = Unmarked.rows() ;
    ostringstream oss; 
    oss << (int)ceil(x/ds) << "x" << (int)ceil(y/ds) << "!";
    Unmarked.resize(oss.str()) ; 
    cout << "Downsampled by " << ds << " to ~" << oss.str() << endl; 
//     Unmarked.write("Img.png") ;    

    Magick2PFM(Unmarked, I) ;

}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//  Create FactoredImage object
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
FactoredImage* GetFactoredImage(const string& input, Params& p) 
{
    FactoredImage* pFI ;
    ////////////////// Compute or load embedding
	PFMImage I ;
	    /////// Read unmarked input image
	    try
	    {
		cout << input << endl ;
		ReadImage(input, I, p.ds) ;
	    }
	    catch (...)
	    {
		cout << " Cannot find File " << input << endl ;
		return NULL;
	    }
	    pFI = new FactoredImage (p, I) ;
    return pFI ;
}
    
    

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//  Parse command line arguments and perform embedding and labelling
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Run(int argc, char* argv[]) 
{
    CmdLine cmd("Euclidean embedding of images", ' ', "0.1");
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////// Parameters for computing embedding
    /////////////////////////////////////////////////////////////
    ValueArg<int> psArg("","patch_size","Patch-size (half-window size)",false,1,"int");
    cmd.add(psArg) ;
    ValueArg<int> dsArg("x","downsample","Downsampling factor",false,1,"int");
    cmd.add(dsArg) ;
    ValueArg<int> nDArg("d","ndims","Number of dims",false,3,"int");
    cmd.add(nDArg) ;
    ///////
    ValueArg<float> NysArg("n","nyssamp","Nystrom subsampling factor",false,0.001,"float");
    cmd.add(NysArg) ;
    ValueArg<float> sigrArg("r","sigr","Sigma_r (range sigma)",false, 1.5,"float");
    cmd.add(sigrArg) ;
    ValueArg<float> sigsArg("s","sigs","Sigma_s (spatial sigma)",false, 1500,"float");
    cmd.add(sigsArg) ;
    ///////
    ValueArg<string> PathArg("p","path","Path to file name. For scribbled-image pairs, give path to folder containing unmarked and marked directories.",false,"../../images/", "string");
    cmd.add(PathArg) ;
    ValueArg<string> FileArg("f","input_file","Input image with path. For scribbled image-pairs, give filename without extension",true,"", "string");
    cmd.add(FileArg) ;
    ValueArg<string> DistMetArg("D","dist","Distance metric",false,"FAB", "{EUC, LAP, BIL, FAB, PAT}");
    cmd.add(DistMetArg) ;
    ///////
    ValueArg<bool> LUVArg("v","luv","Work in LUV colour space",false,true,"bool") ;
    cmd.add(LUVArg) ;
    ///////
    ///////////////////////////////////////////////////////////// Parameters for applications
    MultiArg<string> CRFArg("C","crf","First specify annotation image. Next provide output image name.",false, "string");
    cmd.add(CRFArg) ;   
    MultiArg<double> sigArg("","crf_sigma","Std dev along each dimension for Gaussian weighting.",false, "double");
    cmd.add(sigArg) ;
    MultiArg<double> wtArg("","crf_weight","Weight for factored image.",false, "double");
    cmd.add(wtArg) ;    
/*    ValueArg<int> nlabelsArg("","crf_nlabels","Total num class labels.",false, 21, "double");
    cmd.add(nlabelsArg) ;*/
    ValueArg<string> unaryArg("","crf_unary","Unary potential file.",false, "", "string");
    cmd.add(unaryArg) ;
    
    cmd.parse(argc, argv);
    
    EnumParser<DistMetric> ep ;
    
    cout << "  _____________________ Computing Embedding _____________________ "  << endl ; 
    Params p;
    p.ds = dsArg.getValue() ;
    p.NystromFrac = NysArg.getValue() ;
    p.sigr = sigrArg.getValue() ;
    p.sigs = sigsArg.getValue() ;
    p.nDim = nDArg.getValue() ;
//     p.UseScribbles = UseScribsArg.getValue() ;
//     p.UseShortestPaths = IsomapArg.getValue() ; 
    p.distTYPE  = ep.ParseSomeEnum(DistMetArg.getValue());
    p.LUVSpace = LUVArg.getValue() ;
    p.PatchMetricWSize = psArg.getValue() ;
    cout << p;
    
    string input(FileArg.getValue()) ;
    IMPATH = PathArg.getValue();
    FactoredImage* pFI = GetFactoredImage(input, p) ;
    if (pFI == NULL) return ;
    
    
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// 
    //////  MAP inference in fully-connected CRF    [KK11] 
    vector<string> crfs = CRFArg.getValue() ;
    vector<double> sigs = sigArg.getValue() ;
    vector<double> wts = wtArg.getValue() ;
    double sigma(0.1), wt(10), nlabels(3)  ;
    if (sigs.size()) sigma=sigs[0] ;
    if (wts.size()) wt=wts[0] ;
	    
    if (crfs.size()==3)
    {	
	 for(int ss(0); ss<sigs.size(); ss++) 
		for(int ww(0); ww<wts.size(); ww++)
		{
		    sigma=sigs[ss] ;
		    wt=wts[ww] ;
// 		    cout << " ----------------------- " << sigma << " " << wt << endl;
		    cout << endl << "  _____________________ labelling  using [KK11] _____________________  " << endl << " .. with sig:"  << sigma << " and wt:" << wt << endl;

		    CallCRFVersion(argc, argv, pFI->GetOriginalImage(), pFI, IMPATH+crfs[0], IMPATH+crfs[1], crfs[2], p.ds, sigma, wt, false) ;
		}
    }
    
    /////////////////////////////////////////////////////////////////
	delete pFI ;
}


int main(int argc, char* argv[])
{
    srand(time(NULL)) ;
    
       try
	{
	    Run(argc, argv) ;
	}
	catch(ArgException &e)  // catch any exceptions
	{ cerr << "error: " << e.error() << " for arg " << e.argId() << endl; }	
        
        cout << endl ;
    return 0; 
}
