/*
 * Accurate binary image selection from inaccurate user input
 *
 * Authors: Kartic Subr, Sylvain Paris, Cyril Soler, Jan Kautz
 *
 * Copyright (2013). All rights reserved.
 *
 * For details, see the paper:
 * Accurate binary image selection from inaccurate user input
 * Computer Graphics Forum 32, 2 (2013)
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for educational, research, and non-commercial
 * purposes, without fee and without a signed licensing agreement, is
 * hereby granted, provided that the above copyright notice and this
 * paragraph appear in all copies modifications, and distributions.
 *
 * Any commercial use or any redistribution of this software
 * requires a license.
 *
 * For further details, contact Kartic Subr (kartic@gmail.com)
 */

#ifndef __DIST_HELPER__
#define __DIST_HELPER__

#include <PFMImage.h>
#include <DataMat.h>
#include <map>
#include <string>
#include <Hist.h>

using namespace std; 

typedef enum {EUC, LAP, BIL, FAB, PAT, SSIM, L1, CHISQ, EMD, DMAP, BIH, CORREL} DistMetric; 
extern DistMetric int2DistMetric[15] ; 

////////////////////////////////////////////////
// Parser that reads strings and converts to enum type for distance metric
// Ideally, this could be avoided.
// Todo: Clean up enum usage and use class hierarchy for dissimilarities
////////////////////////////////////////////////
template <typename T>
class EnumParser
{
    map <string, T> enumMap;
public:
    EnumParser();

    T ParseSomeEnum(const string &value)
    { 
	return enumMap[value] ;
    }
};

class FactoredImage ;
////////////////////////////////////////////////
// Convenient helper class that computes distances
// The goal of this class is to abstract the dissimilarity measure used in the embedding
//
// On creation, it computes useful structures such as local histograms
// so that computing distances on the fly is more efficient
// 
////////////////////////////////////////////////
class DistHelper
{
    public:
	DistHelper(const PFMImage& I, DistMetric distTYPE, double sig_r=0.0, double sig_s=0.0, bool LUVSpace=false, int PatchSize=1)  ;
	~DistHelper() ;
	
	double getDistanceSq(int i, int j, int i2, int j2) const ;	

	double getDistance(int i, int j, int i2, int j2) const ;

    private: 
	vector<DataMat> Features ;
	int _nd, _w, _h; 
	double _sigsqr, _sigsqs ;
	DistMetric _dtype ;
	PFMImage _I ;
	
	float c1, c2 ;
	
	DataMat Dfs, Dbs; 
	FactoredImage *_pFI ;
	bool ExistFIs; 
	
	vector<vector<MultivalHist> > _hists ;
	vector<vector<Histogram> > _RGBhists ;
	
	double getDistanceSqNoScribs(int i, int j, int i2, int j2) const ;
};

#endif // __DIST_HELPER__