/*
 * Accurate binary image selection from inaccurate user input
 *
 * Authors: Kartic Subr, Sylvain Paris, Cyril Soler, Jan Kautz
 *
 * Copyright (2013). All rights reserved.
 *
 * For details, see the paper:
 * Accurate binary image selection from inaccurate user input
 * Computer Graphics Forum 32, 2 (2013)
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for educational, research, and non-commercial
 * purposes, without fee and without a signed licensing agreement, is
 * hereby granted, provided that the above copyright notice and this
 * paragraph appear in all copies modifications, and distributions.
 *
 * Any commercial use or any redistribution of this software
 * requires a license.
 *
 * For further details, contact Kartic Subr (kartic@gmail.com)
 */

#ifndef __FACTORED_IMAGE__
#define __FACTORED_IMAGE__

#include <DataMat.h>
#include <PFMImage.h>
#include <DistHelper.h>
#include <vector>
#include <fstream>
#include <map>

using std::pair; 

////////////////////////////////////////////////
// structure containing parameters to embedding
////////////////////////////////////////////////
struct Params
{    
    Params(): ds(1), NystromFrac(0.01), sigr(1.5), sigs(1500), nDim(3), 
			UseScribbles(false), UseShortestPaths(false), distTYPE(EUC), LUVSpace(false), 
			    kNN(512*512), PrecomputeNN (false), knnRad(0.3), nTileX(1), UseXYDims(true), 
				FilterRadius(0.03), Verbosity (2), PatchMetricWSize(1)  {}
    Params(const string& Fname) {ifstream ifs(Fname.c_str()); if(ifs) Read(ifs); ifs.close() ;}
    
    int ds ;
    double NystromFrac ;
    double sigr, sigs ;
    int nDim;
    bool UseScribbles;
    bool UseShortestPaths; 
    DistMetric distTYPE  ;
    bool LUVSpace, PrecomputeNN ;
    int kNN; 
    double knnRad;
    int nTileX ;
    bool UseXYDims ;
    double FilterRadius ;
    int Verbosity ;
    int PatchMetricWSize ;
    
    friend ostream& operator<<(ostream& os, const Params& p) ;
    friend istream& operator>>(istream& is, Params& p) ;
    
    void Write(ofstream& ofs) const      {  ofs << *this << endl ;  	cout << " Wrote params: " << *this << endl ;     }   
    void Read(ifstream& ifs)       { 	ifs >> *this; 	cout << " Read params: " << *this << endl ;    }
};


////////////////////////////////////////////////
//  Abstraction for image embedding 
// 
//  Created using embedding params and an input image 
//  Represents an embedding of all image pixels 
//  in a feature space so that the Euclidean distance between the feature vectors of pixels
//  is approximate equal to a specified dissimilarity measure (params.distTYPE) 
// 
//  the operator (i,j) then returns the feature vector corresponding to the (i,j)th pixel
////////////////////////////////////////////////
class FactoredImage
{
    public:
	FactoredImage(const Params& param, const PFMImage& I) ;
	FactoredImage(const string& Path) ;
	
	int GetNDims() const {return _O.size(); } 
	int GetWidth() const {return _w;} 
	int GetHeight() const {return _h;} 
	pair<double, double> GetMinMaxCoords () const {return pair<double, double>(_minCoord, _maxCoord);} 
	const vector<double>& GetMinCoordsForDims() const {return _minDims;} 
	const vector<double>& GetMaxCoordsForDims() const {return _maxDims;} 
	
	void ComputeGradientMag(DataMat& ) ;	
	const PFMImage& GetOriginalImage() const {return *_pI;}  
	const vector<double>& GetEigenVals() const {return _Eigvals;}	
	
	double SqDistBetweenPixels(int r1, int c1, int r2, int c2) const ;
	PFMPixel GetPixelColour(int r, int c) const { return (*_pI)(r,c); }
		
	// Use this for dimensions other than 3
	void Cluster(    int nClustersReq, int maxCSize, DataMat& CIds, int& ActualnClusters, const string& Method) ;
		
	vector<double> operator()(int i, int j) const ;
	void Normalize() ;
	void Write (const string& Path) const ;
	
	~FactoredImage() ;
    private:
	vector<double> _Eigvals ;
	vector<DataMat> _O ;
	const PFMImage* _pI ;	
	Params _params; 
	int _w, _h ;
	DistHelper _DH ;
	double _minCoord, _maxCoord, _maxDiam ;
	bool _AllocatedPI ;
	vector<double> _minDims, _maxDims ; // min and max along each dimension
	
	/////////////////////////////////////////////////////////////////////////////////////// Various operations on the image embedding
	void ForceRGBWithoutEmbedding() ;
	void ComputeMinMaxDims()  ;
	void NystromShortestPaths(const vector<double>& d, const vector<int>& ri, const vector<int>& ci, DataMat& C, DataMat& A, vector<int>& P, const int c, const int n) ;
	void NystromRandomSample(vector<int>& P, int c, int n) ;
	void Factorize(const DataMat& C, const DataMat& A, const vector<int>& P, const int n, const int c) ;
	void ComputeSparseDistanceMatrix(vector<int>& ri, vector<int>& ci, vector<double>& d, int & sz) ;
	void NystromAllPairs( DataMat& C, DataMat& A, vector<int>& P, const int c, const int n) ;
	void StoreNN(int k,  int nDim, double rad, int nTileX, bool UseXY) ;
	void StoreNNTile(int k,  int nDim, int OPSpaceDim, int r1, int c1, int r2, int c2, int overlap, int rt1, int ct1, int rt2, int ct2) ;
	void StoreNNTileCPU(double rad, int MaxNbrs,  int nDim, bool UseXY, int r1, int c1, int r2, int c2, int overlap, int rt1, int ct1, int rt2, int ct2) ;
	void ComputeDiffusionMap(const Params& param) ;
	void HDClusterCudaEM(int nSeeds, DataMat& CIds, int& nClusters) ;
	void HDClusterCudaKMeans(double cTol, int kmeansK, DataMat& CIds, int& nClusters) ;
};

#endif // __FACTORED_IMAGE__