/*
 * Accurate binary image selection from inaccurate user input
 *
 * Authors: Kartic Subr, Sylvain Paris, Cyril Soler, Jan Kautz
 *
 * Copyright (2013). All rights reserved.
 *
 * For details, see the paper:
 * Accurate binary image selection from inaccurate user input
 * Computer Graphics Forum 32, 2 (2013)
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for educational, research, and non-commercial
 * purposes, without fee and without a signed licensing agreement, is
 * hereby granted, provided that the above copyright notice and this
 * paragraph appear in all copies modifications, and distributions.
 *
 * Any commercial use or any redistribution of this software
 * requires a license.
 *
 * For further details, contact Kartic Subr (kartic@gmail.com)
 */

#ifndef _PFMIMAGE_H_
#define _PFMIMAGE_H_

#include <stdlib.h>
#include <iostream>
#include <cstdio>
#include <cmath>
#include <string>
#include <iostream>

typedef float PFMChannel;  
using std::string;
using std::ostream;

////////////////////////////////////////////////
// Pixels in PFMImage
////////////////////////////////////////////////
struct PFMPixel {
	PFMPixel() { r = 0; g = 0; b = 0; }
	PFMPixel( PFMChannel r_, PFMChannel g_, PFMChannel b_ ) { r = r_; g = g_; b = b_; }
	PFMPixel& operator += (const PFMPixel B) { r+=fabs(B.r); g+=fabs(B.g); b+=fabs(B.b); return *this; }
	const PFMPixel& operator = (const PFMPixel B) { r=fabs(B.r); g=fabs(B.g); b=fabs(B.b); return *this ;}
	double pix2d() const {return .15*r+.7*g+.15*b; }
        PFMChannel r;
	PFMChannel g;
	PFMChannel b;        
};

inline ostream& operator<<(ostream&os, const PFMPixel& p)
{
    os  << p.r << " " << p.g << " " << p.b << " ";
    return os ;
}

////////////////////////////////////////////////
//  Computed L component values are in the range [0 to 100].
//  Computed U component values are in the range [-124 to 220].
//  Computed V component values are in the range [-140 to 116].
// return values in [0,1]
////////////////////////////////////////////////
inline PFMPixel RGB2LUV(const PFMPixel& p)
{
    double X = 0.412453*p.r + 0.35758 *p.g + 0.180423*p.b;
    double Y = 0.212671*p.r + 0.71516 *p.g + 0.072169*p.b;
    double Z = 0.019334*p.r + 0.119193*p.g + 0.950227*p.b;
    static const double xn = 0.312713 ;
    static const double yn = 0.329016;
    static const double  Yn = 1.0 ;
    static const double un = 4*xn / (-2*xn + 12*yn + 3);
    static const double vn = 9*yn / (-2*xn + 12*yn + 3);
    double denom = (X + 15*Y + 3*Z) ;
    if (fabs(denom)<1e-8) return PFMPixel(1,1,1) ;
    double u = 4*X / denom;////////////////////////////////////////////////

    double v = 9*Y / denom;
    double L = 116 * pow((Y/Yn), (1/3)) - 16;
    double U = 13*L*(u-un);
    double V = 13*L*(v-vn);
//     std::cout << p << " ---------- " << PFMPixel(L/100, (U+134)/354, (V+140)/256)   << "\t" << std::flush ;
//     std::cin.ignore() ;
    return PFMPixel(L/100, (U+134)/354, (V+140)/256) ;
}

inline PFMPixel operator*(double s, const PFMPixel& B)
{return PFMPixel(s*B.r, s*B.g, s*B.b) ;}

inline PFMPixel operator+(const PFMPixel& A, const PFMPixel& B)
{return PFMPixel(A.r+B.r, A.g+B.g, A.b+B.b) ;}

inline PFMPixel operator*(const PFMPixel& A, const PFMPixel& B)
{return PFMPixel(A.r*B.r, A.g*B.g, A.b*B.b) ;}

inline PFMPixel operator-(const PFMPixel& A, const PFMPixel& B)
{return PFMPixel(A.r-B.r, A.g-B.g, A.b-B.b) ;}

typedef union { unsigned char b[4] ; PFMChannel val; } UFUnion ;

////////////////////////////////////////////////
// Internal image data structure
////////////////////////////////////////////////
struct PFMImage 
{	
        public:
               PFMImage( int x_res=0, int y_res=0):_AvgApproxPower (0), _MaxApproxPower(0), _TotalApproxPower(0) 
                {
                        width  = x_res;
                        height = y_res;
                        if (width+height!=0)
                                pixels = new PFMPixel[ width * height ];
                        PFMPixel *p = pixels;
                        for( int i = 0; i < width * height; i++ ) *p++ = PFMPixel(0.0,0.0,0.0);
                }
                
                PFMImage( const PFMImage& r):_AvgApproxPower (0), _MaxApproxPower(0), _TotalApproxPower(0) 
                {
// 			std::cout << "Calling copy constructor of PFMImage" << std::endl ;
                        width  = r.width;
                        height = r.height; 
                        if (width+height!=0)
                                pixels = new PFMPixel[ width * height ];
                        PFMPixel *p = pixels;
                        PFMPixel *q = r.pixels;
			for( int i = 0; i < width * height; i++ ) *p++ = *q++;
                }
                
               inline void Resize( int width, int height ) ;
               inline bool Open( const char *file_name );
               inline ~PFMImage() { if(width+height!=0) delete[] pixels; }
	       inline PFMPixel &operator()( int i, int j ) {  return *( pixels + ( i * width + j ) ); }  
	       inline const PFMPixel &operator()( int i, int j ) const { return *( pixels + ( i * width + j ) ); }  
	       inline PFMPixel &operator()( int idx ) {  return *( pixels + idx ); }  
	       inline const PFMPixel &operator()( int idx ) const { return *( pixels + idx ); }  
        	
                inline double ApproxPower( int i, int j ) const 
                { if ( (i>height) || (j>width) ) {std::cerr << "Overread .." << std::endl ; exit(1);} const PFMPixel p = *( pixels + ( i * width + j ) ); return p.r+p.g+p.b; }          
                
//             inline void AntialiasedPixelDraw( double i, double j, const Color& C) ;
                inline void Splat( int i, int j, int SplatDiam, double Sig, PFMPixel P) ;
	       PFMPixel *pixels;
	       int    width;
	       int    height;
               mutable string _ID ;
               double _MaxApproxPower, _AvgApproxPower, _TotalApproxPower ;        
        
        private:
                inline PFMChannel Build(bool CorrectEndian, const UFUnion& u);
};

inline void PFMImage::Resize( int w, int h) 
{
        if (pixels!=NULL && width+height!=0)  delete [] pixels; 
        width = w ;
        height = h ;
//         std::cout << "Resizing "<< std::endl ;
        pixels = new PFMPixel[ width*height ] ;
//         std::cin.ignore();         
}

#endif
